/* 
   orxonox - the future of 3D-vertical-scrollers

   Copyright (C) 2004 orx

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   ### File Specific:
   main-programmer: Benjamin Grauer
   co-programmer: ...
*/

#include "framework.h"

#include "physics_engine.h"
#include "particle_engine.h"

#include "gui_gtk.h"
#include "p_node.h"
#include "null_parent.h"
#include "state.h"
#include "debug.h"
#include "graphics_engine.h"
#include "light.h"
#include "resource_manager.h"
#include "camera.h"

#include <pthread.h>

int verbose;

pthread_mutex_t mutex;



void Framework::initModule()
{
  // Creating a Test Particle System
  ParticleSystem* system = new ParticleSystem(100000, PARTICLE_SPRITE);
  system->setLifeSpan(5);
  system->setConserve(.999);
  system->setRadius(4, 3, 1, 2);
  system->setColor(.5,0,0,.5, 1,1,0,1, 0,0,0,0);

  // Creating a Test Particle Emitter
  ParticleEmitter* emitter = new ParticleEmitter(Vector(0 , 1, 0), M_PI_4, 20, .1);
  emitter->setType(EMITTER_DOT  );
  emitter->setSize(0);
  emitter->setRelCoor(Vector(0,0,0));
  
  // Add the Flow from the Emitter into the System
  ParticleEngine::getInstance()->addConnection(emitter, system);
}

void* Framework::mainLoop(void* tmp)
{
  Framework* framework = Framework::getInstance();
  while(true)
    {
      // keyhandler returns false if sdl gets quit by some event
      if (!framework->keyHandler())
	pthread_exit(NULL);

      // tick the scene
      float dt = framework->tick();

      // Draw the scene
      pthread_mutex_lock (&mutex); 
      framework->draw(dt);
      pthread_mutex_unlock (&mutex);
    }
}


bool Framework::draw(float dt)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  glLoadIdentity(); // Reset the view
  
  ParticleEngine::getInstance()->draw(dt);

  camera->apply();
  
  SDL_GL_SwapBuffers(); // Swap the buffers
}


float Framework::tick()
{
  currFrame = SDL_GetTicks();
  float dt = (float)(currFrame - lastFrame) / 1000.0;
  lastFrame = currFrame;

  ParticleEngine::getInstance()->tick(dt);

  NullParent::getInstance()->update(dt);
  return dt;
}


bool Framework::keyHandler()
{
  // This is the main loop for the entire program and it will run until done==TRUE
  {
    // And poll for events
    SDL_Event event;
    while(SDL_PollEvent(&event))
    {
      switch (event.type) {
      case SDL_MOUSEMOTION:
	{
	  Vector view = camera->getTarget()->getAbsCoor() - camera->getAbsCoor();
	  Vector up = Vector(0, 1, 0);
	  up = camera->getAbsDir().apply(up);
	  Vector h = up.cross(view);
	  Vector v = h.cross(view);
	  h.normalize();
	  v.normalize();

	  if (mouseDown[1])
	    camera->setRelCoor(camera->getRelCoor()+ h * event.motion.xrel *.05 - v * event.motion.yrel * .05);
	  if (mouseDown[3])
	    camera->getTarget()->setRelCoor(camera->getTarget()->getRelCoor()+ h * event.motion.xrel *.05 - v * event.motion.yrel * .05);
	    
	}
	break;
      case SDL_MOUSEBUTTONDOWN:
	switch (event.button.button)
	  {
	  case 4:
	    PRINTF(4)("MouseWheel up\n");
	    camera->setRelCoor(camera->getRelCoor() + (camera->getTarget()->getAbsCoor() - camera->getAbsCoor())*.1);
	    break;
	  case 5:
	    PRINTF(4)("MouseWheel down\n");
	    camera->setRelCoor(camera->getRelCoor() - (camera->getTarget()->getAbsCoor() - camera->getAbsCoor())*.1);
	    break;
	  case 1:
	  case 2:
	  case 3:
	    mouseDown[event.button.button] = true;
	    break;
	  }
	    
	break;
      case SDL_MOUSEBUTTONUP:
	switch (event.button.button)
	  {
	  case 1:
	  case 2:
	  case 3:
	    mouseDown[event.button.button] = false;
	    break;
	  }
	break;
      case SDL_VIDEORESIZE:
	GraphicsEngine::getInstance()->resolutionChanged(&event.resize);
	break;
      case SDL_KEYDOWN:
	switch (event.key.keysym.sym)
	  {
	  case SDLK_x:
	    
	    break;
	  case SDLK_c:

	    break;
	  case SDLK_i:
	    ParticleEngine::getInstance()->debug();
	    break;
	  case SDLK_a:
	    camera->setRelCoor(camera->getRelCoor() + (camera->getTarget()->getAbsCoor() - camera->getAbsCoor())*.1);
	    break;
	  case SDLK_z:
	    camera->setRelCoor(camera->getRelCoor() - (camera->getTarget()->getAbsCoor() - camera->getAbsCoor())*.1);
	    break;
	  case SDLK_h:
	    this->printHelp();
	    break;
	  }
	break;
	    
        // If a quit event was recieved
      case SDL_QUIT:
	// then we're done and we'll end this program
	return false;
	break;
      default:
	break;
      }

    }

    // Get the state of the keyboard keys
    keys = SDL_GetKeyState(NULL);

    // and check if ESCAPE has been pressed. If so then quit
    if(keys[SDLK_ESCAPE]) return false;
  }
  return true;
}

Framework* Framework::singletonRef = NULL;

Framework* Framework::getInstance(void)
{
  if (Framework::singletonRef == NULL)
    Framework::singletonRef = new Framework();
  return Framework::singletonRef;
}

Framework::Framework()
{
  this->lastFrame = 0;
  // Create a new OpenGL window with the title "Cone3D Basecode" at
  // 640x480x32, fullscreen and check for errors along the way
  GraphicsEngine::getInstance();

  LightManager::getInstance();
  glEnable(GL_TEXTURE_2D);

  // Build the font from a TGA image font.tga in the data directory
  // Hide the mouse cursor
  SDL_ShowCursor(2);

  for (int i = 0; i <MOUSE_BUTTON_COUNT; i++)
    mouseDown[i] = false;

  ResourceManager::getInstance()->setDataDir(DATA_DIRECTORY);
 

  initModule();

  camera = new Camera();

  State::getInstance()->setCamera(camera, camera->getTarget());

  camera->setAbsCoor(Vector(10, 10, 0));

}

Framework::~Framework()
{
  delete GraphicsEngine::getInstance();

}



void Framework::printHelp(void) const
{
  PRINT(0)(" Help for the frameWork\n");
  PRINT(0)("========================\n");
  PRINT(0)("h - print thisHelp\n");
  PRINT(0)("i - state Information\n\n");
  PRINT(0)("\n");


}

int emitterChange(GtkWidget* nonInterest, void* widget)
{
  Option* option = (Option*) widget;
  const char* name = option->getTitle();
  char* valueC = option->save();
  float value = atof(valueC);

  ParticleEmitter* tmpEmit = ParticleEngine::getInstance()->getEmitterByNumber(1);
  if (tmpEmit)
    {
      if (!strcmp(name, "EmissionRate"))
	{
	  tmpEmit->setEmissionRate(value);
	  PRINT(3)("EmissionRate set to %f\n", value);
	}
      else if (!strcmp(name, "Velocity"))
	{
	  tmpEmit->setEmissionVelocity(value);
	  PRINT(3)("Velocity set to %f\n", value);
	}
      else if(!strcmp(name, "SpreadAngle"))
	{
	  tmpEmit->setSpread(value);
	  PRINT(3)("SpreadAngle set to %f\n", value);
	}
      else if(!strcmp(name, "EmitterType"))
	{
	  if (!strcmp(valueC, "EMITTER_DOT"))
	    tmpEmit->setType(EMITTER_DOT);
	  else if (!strcmp(valueC, "EMITTER_PLANE"))
	    tmpEmit->setType(EMITTER_PLANE);
	  else if (!strcmp(valueC, "EMITTER_CUBE"))
	    tmpEmit->setType(EMITTER_CUBE);
	  PRINT(3)("EmitterType set to %s\n", valueC);	  
	}
      else if(!strcmp(name, "EmitterSize"))
	{
	  tmpEmit->setSize(value);
	  PRINT(3)("EmitterSize set to %f\n", value);
	}
    }
  delete valueC;
}


int systemChange(GtkWidget* nonInterest, void* widget)
{
  Option* option = (Option*) widget;
  const char* name = option->getTitle();
  char* valueC = option->save();
  float value = atof(valueC);

  ParticleSystem* tmpSys = ParticleEngine::getInstance()->getSystemByNumber(1);
  if (tmpSys)
    {
      if (!strcmp(name, "StartRadius"))
	{
	  tmpSys->setRadius(value, tmpSys->getEndRadius());
	  PRINT(3)("ParticleStartRadius set to %f\n", value);
	}
      else if (!strcmp(name, "EndRadius"))
	{
	  tmpSys->setRadius( tmpSys->getStartRadius(), value);
	  PRINT(3)("ParticleEndRadius set to %f\n", value);
	}

      else if (!strcmp(name, "LifeSpan"))
	{
	  tmpSys->setLifeSpan(value);
	  PRINT(3)("ParticleLifeSpan set to %f\n", value);
	}

      else if (!strcmp(name, "ConserveFactor"))
	{
	  tmpSys->setConserve(value);
	  PRINT(3)("ParticleConserveFactor set to %f\n", value);
	}

      else if (!strcmp(name, "ParticleType"))
	{
	  if (!strcmp(valueC, "PARTICLE_DOT"))
	    tmpSys->setType(PARTICLE_DOT);
	  else if (!strcmp(valueC, "PARTICLE_SPARK"))
	    tmpSys->setType(PARTICLE_SPARK);
	  else if (!strcmp(valueC, "PARTICLE_SPRITE"))
	    tmpSys->setType(PARTICLE_SPRITE);

	  PRINT(3)("ParticleType set to %s\n", valueC);
	}

      else if (!strcmp(name, "InheritSpeed"))
	{
	  tmpSys->setInheritSpeed(value);
	  PRINT(3)("ParticleInheritSpeed set to %f\n", value);
	}

    }
  delete valueC;
}


int quitGui(GtkWidget* widget, void* data)
{
#ifdef HAVE_GTK2
  gtk_main_quit();
  while(gtk_events_pending()) gtk_main_iteration();
#endif /* HAVE_GTK2 */
}

void* Framework::initGui(void* argv)
{
  Window* guiMainWindow = NULL;
  
  initGUI(0, NULL);
  
  guiMainWindow = new Window("ParticlesFUN");
  {
    Box* windowBox = new Box('v');
    {
      Frame* emitterFrame = new Frame("emitter-settings");
      {
	Box* emitterBox = new Box('v');
	{
	  emitterBox->fill(new Label("EmissionRate"));
	  Slider* EmissionRate = new Slider("EmissionRate", 0, 1000);
	  EmissionRate->connectSignal("value_changed", (void*)EmissionRate, emitterChange );
	  emitterBox->fill(EmissionRate);
	  
	  emitterBox->fill(new Label("Velocity"));
	  Slider* velocity = new Slider("Velocity", 0, 2);
	  velocity->setExactness(3);
	  velocity->connectSignal("value_changed", (void*)velocity, emitterChange );
	  emitterBox->fill(velocity);
	  
	  emitterBox->fill(new Label("SpreadAngle"));
	  Slider* SpreadAngle = new Slider("SpreadAngle", 0, M_PI);
	  SpreadAngle->setExactness(3);
	  SpreadAngle->connectSignal("value_changed", (void*)SpreadAngle, emitterChange );
	  emitterBox->fill(SpreadAngle);

	  emitterBox->fill(new Label("EmitterType"));
	  Menu* EmitterType = new Menu("EmitterType");
	  EmitterType->addItem("EMITTER_DOT");
	  EmitterType->addItem("EMITTER_PLANE");
	  EmitterType->addItem("EMITTER_CUBE");
	  EmitterType->connectSignal("changed", (void*)EmitterType, emitterChange );	  
	  emitterBox->fill(EmitterType);

	  emitterBox->fill(new Label("EmitterSize"));
	  Slider* EmitterSize = new Slider("EmitterSize", 0, 100);
	  EmitterSize->setExactness(1);
	  EmitterSize->connectSignal("value_changed", (void*)EmitterSize, emitterChange );
	  emitterBox->fill(EmitterSize);
	}
	emitterFrame->fill(emitterBox);
      }
      windowBox->fill(emitterFrame);
      
      Frame* systemFrame = new Frame("system-settings");
      {
	Box* systemBox = new Box('v');
	{
	  systemBox->fill(new Label("StartRadius"));
	  Slider* StartRadius = new Slider("StartRadius", 0, 10);
	  StartRadius->setExactness(3);
	  StartRadius->connectSignal("value_changed", (void*)StartRadius, systemChange );
	  systemBox->fill(StartRadius);

	  systemBox->fill(new Label("EndRadius"));
	  Slider* EndRadius = new Slider("EndRadius", 0, 10);
	  EndRadius->setExactness(3);
	  EndRadius->connectSignal("value_changed", (void*)EndRadius, systemChange );
	  systemBox->fill(EndRadius);

	  systemBox->fill(new Label("LifeSpan"));
	  Slider* LifeSpan = new Slider("LifeSpan", 0, 10);
	  LifeSpan->setExactness(3);
	  LifeSpan->connectSignal("value_changed", (void*)LifeSpan, systemChange );
	  systemBox->fill(LifeSpan);
	  
	  systemBox->fill(new Label("ConserveFactor"));
	  Slider* ConserveFactor = new Slider("ConserveFactor", 0, 1);
	  ConserveFactor->setExactness(3);
	  ConserveFactor->connectSignal("value_changed", (void*)ConserveFactor, systemChange );
	  systemBox->fill(ConserveFactor);

	  systemBox->fill(new Label("ParticleType"));
	  Menu* ParticleType = new Menu("ParticleType");
	  ParticleType->addItem("PARTICLE_DOT");
	  ParticleType->addItem("PARTICLE_SPARK");
	  ParticleType->addItem("PARTICLE_SPRITE");
	  ParticleType->connectSignal("changed", (void*)ParticleType, systemChange );	  
	  systemBox->fill(ParticleType);
	  
	  systemBox->fill(new Label("InheritSpeed"));
	  Slider* InheritSpeed = new Slider("InheritSpeed", 0, 1);
	  InheritSpeed->setExactness(3);
	  InheritSpeed->connectSignal("value_changed", (void*)InheritSpeed, systemChange );
	  systemBox->fill(InheritSpeed);


	}
	systemFrame->fill(systemBox);
      }
      windowBox->fill(systemFrame);
      
      Button* quitButton = new Button("quit");
      
      quitButton->connectSignal("clicked", NULL, quitGui);
      //  Window::mainWindow->connectSignal("remove", this, GuiExec::quitGui);
      Window::mainWindow->connectSignal("destroy", NULL, quitGui);
      
      windowBox->fill(quitButton);
    }
    guiMainWindow->fill(windowBox);
  }
  Window::mainWindow->showall();
  Window::mainWindow->setSize(300, 500);
}

void* Framework::mainloopGui(void* tmp)
{
  mainloopGUI();
}



int main(int argc, char *argv[])
{
  verbose = 3;
  pthread_t guiThread;
  pthread_t openglThread;
  pthread_attr_t attr;
  int status;
  
  pthread_mutex_init(&mutex, NULL);
  
  /* Create threads to perform the dotproduct  */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);




  Framework* framework = Framework::getInstance();

  framework->initGui((void*)argv);
  pthread_create(&guiThread, &attr, Framework::mainloopGui ,(void*) argv);
  //  framework->initGui(argv);

  pthread_create(&openglThread, &attr, Framework::mainLoop, NULL);

  pthread_attr_destroy(&attr);

  pthread_join( guiThread, (void **)&status);	
  pthread_join( openglThread, (void **)&status);	

  delete framework;
  // Kill the GL & SDL screens
  // And quit
  return 0;
}
